@@ -17,16 +17,17 @@
 """
 from __future__ import annotations
 
+import asyncio
 from asyncio import get_running_loop, AbstractEventLoop, Queue as AsyncioQueue
 from concurrent.futures import ThreadPoolExecutor
 from datetime import datetime
+from functools import partial
 from logging import LogRecord, Handler, NOTSET, ERROR
-from queue import Queue
-from threading import Thread, Event
-from typing import List, Any
+from typing import List, Awaitable
 
 from aiohttp import ClientSession
-from discord import SyncWebhook, Webhook, Embed
+from discord import SyncWebhook, Webhook, Embed, Message
+from discord.utils import MISSING
 
 DEPENDENCY_LOGGERS: set[str] = {
     "aiohttp",
@@ -43,14 +44,23 @@ def filter_out_dependencies(record: LogRecord) -> int:
     Filters out records from dependencies. This is needed to not have an infinite loop of logging.
     """
 
-    should_filter: bool = any(
-        [record.name.startswith(dependency) for dependency in DEPENDENCY_LOGGERS]
-    )
+    should_filter: bool = any([record.name.startswith(dependency) for dependency in DEPENDENCY_LOGGERS])
 
     return 0 if should_filter else 1
 
 
-class DiscordWebhookHandler(Handler, Thread):  # type: ignore  # not a catostrophic clash of DiscordWebhookHandler.name
+def loop_or_none() -> AbstractEventLoop | None:
+    """
+    Gets the current running loop, or None if there is no running loop.
+    """
+
+    try:
+        return get_running_loop()
+    except RuntimeError:
+        return None
+
+
+class DiscordWebhookHandler(Handler):
     """A logging handler that allows you to send messages with Discord Webhooks."""
 
     def __init__(
@@ -59,7 +69,7 @@ class DiscordWebhookHandler(Handler, Thread):  # type: ignore  # not a catostrop
         level: int = NOTSET,
         *,
         can_format: bool = False,
-        run_async: bool = False,
+        event_loop: AbstractEventLoop | None = None,
         text_send_on_error: str | None = None,
         text_send_on_error_threshold: int = ERROR,
     ) -> None:
@@ -68,19 +78,19 @@ class DiscordWebhookHandler(Handler, Thread):  # type: ignore  # not a catostrop
         :param level: The level to log at.
         :param can_format: Whether to format the message using the formatter, or to just send it in the Discord
         description.
-        :param run_async: Whether to run the webhook in an async context. This is useful if your app is async. If your
-        app isn't async (it is using threading or another concurrency method not based on asyncio), then you should
-        install the `sync` extra or install your own version of `requests`.
+        :param event_loop: The event loop to use. If this is None, then no event loop will be used, and a thread will instead
+        be used to send messages. This is useful if you are using a synchronous Discord library. Download the `sync`
+        extra to use this without extra dependency handling.
         :param text_send_on_error: The text to send if an error occurs while sending a message. If this is None, then
         no text will be sent on error. This is useful to send stuff like a mention (<@440468612365680650>).
         :param text_send_on_error_threshold: The level at which to send the text on error. Defaults to logging.ERROR.
         """
-        Handler.__init__(self, level)
+        super().__init__(level)
 
-        self.addFilter(filter_out_dependencies)
+        self.addFilter(filter_out_dependencies)  # type: ignore  # mypy doesn't like this
 
         self._format: bool = can_format
-        self._async: bool = run_async
+        self._async: bool = event_loop is not None and isinstance(event_loop, AbstractEventLoop)
 
         self._aiohttp_session: ClientSession | None = None
 
@@ -97,98 +107,88 @@ class DiscordWebhookHandler(Handler, Thread):  # type: ignore  # not a catostrop
             self._webhook = webhook
             self._aiohttp_session = webhook.session
 
-        self._queue: AsyncioQueue[tuple[tuple[Any, ...], dict[str, Any]]] | Queue[
-            tuple[tuple[Any, ...], dict[str, Any]]
-        ]
-        self._shutdown_event: Event | None = None
-        if self._async:
+        self._runner: AbstractEventLoop | ThreadPoolExecutor
+        self._queue: AsyncioQueue[tuple[LogRecord | None, Awaitable[Message]]] | None = None
+        if event_loop is not None:
+            self._runner = event_loop
             self._queue = AsyncioQueue()
         else:
-            self._queue = Queue()
-            self._shutdown_event = Event()
-            Thread.__init__(self, name=self.name, daemon=True)
-            self.start()
+            self._runner = ThreadPoolExecutor(max_workers=1, thread_name_prefix="dislog")
 
         self._text_send_on_error = text_send_on_error
         self._text_send_on_error_threshold = text_send_on_error_threshold
 
-    def run(self) -> None:
-        assert not self._async, "Cannot run a thread in an async context."
+    async def _async_consume(self) -> None:
+        assert isinstance(self._queue, AsyncioQueue), "Queue is not an asyncio.Queue!"
 
-        assert self._shutdown_event is not None, "Shutdown event is None"
-        assert isinstance(self._queue, Queue), "Queue is not a queue.Queue"
-
-        while not self._shutdown_event.is_set() or not self._queue.empty():
-            args, kwargs = self._queue.get()
-            self._webhook.send(*args, **kwargs)
+        record, coro = await self._queue.get()
+        try:
+            await coro
+        except Exception:  # noqa
+            if record is not None:
+                self.handleError(record)
+        finally:
             self._queue.task_done()
 
-    async def _send_one_async(self) -> None:
-        assert isinstance(self._queue, AsyncioQueue), "Queue is not an asyncio.Queue"
-        assert isinstance(self._webhook, Webhook), "Webhook is not a discord.Webhook"
-
-        args, kwargs = self._queue.get_nowait()
-        await self._webhook.send(*args, **kwargs)
-
-    def _send(self, *args, **kwargs) -> None:
-        self._queue.put_nowait((args, kwargs))
+    def _send(self, record: LogRecord | None, *args, **kwargs) -> None:
+        send_partial = partial(self._webhook.send, *args, **kwargs)
         if self._async:
-            assert self._aiohttp_session is not None, "Session is None"
-            loop: AbstractEventLoop
-            if (
-                getattr(self._aiohttp_session, "loop", None) is None
-            ):  # loop is deprecated
-                loop = get_running_loop()
-            else:
-                loop = self._aiohttp_session.loop
+            assert isinstance(self._runner, AbstractEventLoop), "Runner is not an event loop!"
+            assert isinstance(self._webhook, Webhook), "Webhook is not a discord.Webhook"
+            assert isinstance(self._queue, AsyncioQueue), "Queue is not an asyncio.Queue!"
 
-            loop.create_task(self._send_one_async())
-            # note: this might not happen instantly. that's ok though, because we timestamp the message.
+            send_coro: Awaitable[Message] = send_partial()  # type: ignore  # mypy doesn't like this
+            tup = (record, send_coro)
+            self._queue.put_nowait(tup)
+
+            consume_coro = self._async_consume()
+
+            # This needs to be a callback so that the shield runs on the same thread as the EventLoop
+            self._runner.call_soon_threadsafe(partial(asyncio.shield, consume_coro))
+        else:
+            assert isinstance(self._runner, ThreadPoolExecutor), "Runner is not a thread pool!"
+            assert isinstance(self._webhook, SyncWebhook), "Webhook is not a discord.SyncWebhook"
+
+            self._runner.submit(send_partial)
 
     def _close_sync(self) -> None:
-        assert isinstance(self._queue, Queue), "Queue is not a queue.Queue"
-        assert self._shutdown_event is not None, "Shutdown event is None"
-        assert isinstance(
-            self._webhook, SyncWebhook
-        ), "Webhook is not a discord.SyncWebhook"
+        assert isinstance(self._runner, ThreadPoolExecutor), "Runner is not a thread pool!"
+        assert isinstance(self._webhook, SyncWebhook), "Webhook is not a discord.SyncWebhook"
 
-        self._shutdown_event.set()
-        self._send("**CLOSED**")
-        self._queue.join()
+        self._runner.shutdown(wait=True)
 
     async def _close_async(self) -> None:
         assert isinstance(self._queue, AsyncioQueue), "Queue is not an asyncio.Queue"
         assert isinstance(self._webhook, Webhook), "Webhook is not a discord.Webhook"
         assert self._aiohttp_session is not None, "Session is None"
 
-        self._send("**CLOSED**")
         await self._queue.join()  # noqa  # pycharm is wrong
         await self._aiohttp_session.close()
 
     def close(self) -> None:
+        self._send(None, "**CLOSED**")
         if self._async:
-            assert isinstance(
-                self._queue, AsyncioQueue
-            ), "Queue is not an asyncio.Queue"
-            assert isinstance(
-                self._webhook, Webhook
-            ), "Webhook is not a discord.Webhook"
-
+            assert isinstance(self._runner, AbstractEventLoop), "Runner is not an event loop!"
+            assert isinstance(self._webhook, Webhook), "Webhook is not a discord.Webhook"
             assert self._aiohttp_session is not None, "Session is None"
-            loop: AbstractEventLoop
-            if (
-                getattr(self._aiohttp_session, "loop", None) is None
-            ):  # loop is deprecated
-                loop = get_running_loop()
-            else:
-                loop = self._aiohttp_session.loop
 
-            if loop.is_running():
-                loop.create_task(self._close_async())
-            elif not loop.is_closed():
-                loop.run_until_complete(self._close_async())
-            else:
-                raise RuntimeError("Cannot run async without an event loop candidate!")
+            cancel_task = self._runner.create_task(self._close_async())
+
+            if self._runner.is_closed():
+                raise RuntimeError("Event loop is closed, cannot close handler!")
+
+            running_loop = loop_or_none()
+            if not self._runner.is_running():
+                # The loop isn't running, but it's not closed.
+                self._runner.run_until_complete(cancel_task)
+            elif running_loop is self._runner:
+                # The loop is running, and it's the current loop.
+                # There is nothing we can do that will not deadlock.
+                pass
+            elif running_loop is None:
+                # The loop is running, but it's not the current loop.
+                concurrent_future = asyncio.run_coroutine_threadsafe(cancel_task, self._runner)
+                concurrent_future.result()  # We can wait.
         else:
             self._close_sync()
         super().close()
@@ -216,14 +216,15 @@ class DiscordWebhookHandler(Handler, Thread):  # type: ignore  # not a catostrop
             }
         )
 
+        content: str | MISSING = MISSING
         if (
             self._text_send_on_error_threshold <= record.levelno
             and self._text_send_on_error is not None
             and len(self._text_send_on_error) > 0
         ):
-            self._send(self._text_send_on_error, embeds=[embed])
-        else:
-            self._send(embeds=[embed])
+            content = self._text_send_on_error
+
+        self._send(record, embed=embed, content=content)
 
 
 __all__: List[str] = ["DiscordWebhookHandler"]
